# Copyright (C) 2010-2012 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Signature (XML-DSIG) support."""

from util import get_frame_globals


class SignError(Exception):
  """an error during a signature request."""


class VerifyError(Exception):
  """an error during the verfication of a signature."""



class Signable(object):
  """Abstract mixin class supporting signing.

  Defines method ``request_signature`` which calls for signature
  creation in the ``toxml`` method.

  Note: requires ``Signable`` to be inherited before the ``pyxb`` base
  class defining ``toxml``.

  Class level configuration:

    S_ID_ATTRIBUTE
      attribute name used to determine the signature reference uri
      (default: ``'ID'``)

    S_SIGNATURE_ATTRIBUTE
      attribute name indicating the signature element to be used for signing
      (default: ``'Signature'``)

    S_GET_KEYNAME
      method to determine the keyname to used for signing.
  """

  S_ID_ATTRIBUTE = 'ID'
  S_SIGNATURE_ATTRIBUTE = 'Signature'
  S_GET_KEYNAME = None

  __signature_request = None


  def request_signature(self, keyname=None, context=None):
    """request signature in a following ``toxml``.

    *keyname* identifies the key to be used. If ``None``, it is
    determined by ``S_GET_KEYNAME``.

    *context* determines the signature context. If ``None``,
    ``default_sign_context`` is used.
    """

    self.__signature_request = _SignatureRequest(keyname, context)


  # this does not work -- must work at "DOM" level
  def _toDOM_csc(self, dom_support, parent):
    request = self.__signature_request
    if request is None:
      return super(Signable, self)._toDOM_csc(dom_support, parent)
    # prepare adding signature template
    sa = self.S_SIGNATURE_ATTRIBUTE
    s = getattr(self, sa)
    assert s is None, "signature attribute not None"
    # extend ``dom_support`` class, if necessary
    if not isinstance(dom_support, _BindingDOMSupportSignatureExtension):
      cls = dom_support.__class__
      dom_support.__class__ = type(cls)(
        cls.__name__,
        (_BindingDOMSupportSignatureExtension, cls),
        {}
        )
    dom_support.add_signature_request(request, self)
    try:
      # build the template
      from pyxb.bundles.wssplat.ds import Signature, SignedInfo, \
           CanonicalizationMethod, SignatureMethod, \
           Reference, Transforms, Transform, DigestMethod, DigestValue, \
           SignatureValue
      setattr(
        self, sa,
        Signature(
          SignedInfo(
            CanonicalizationMethod(Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"),
            SignatureMethod(Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"),
            Reference(
              Transforms(
                Transform(Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"),
                Transform(Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"),
                ),
              DigestMethod(Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"),
              DigestValue(''),
              URI=u'#' + getattr(self, self.S_ID_ATTRIBUTE),
              ),
            ),
          SignatureValue(''),
          )
        )
      return super(Signable, self)._toDOM_csc(dom_support, parent)
    finally:
      setattr(self, sa, None) # restore signature attribute

  ###########################################################################
  # verification support
  #   SAML2 allows signatures on enclosing elements to verify its
  #   subelements as well. Even a transport verification can verify
  #   an object.
  # For the moment, we support signatures only by the issuing authority,
  #  identified by its keyname
  __verifying_keyname = None

  def set_signature_verification(self, keyname):
    """state that this object has been verified by *keyname*."""
    self.__verifying_keyname = keyname

  def verified_signature(self):
    """return true when this object has been verified."""
    return self.__verifying_keyname == self.S_GET_KEYNAME()



class _SignatureRequest(object):
  """auxiliary class to manage signature information."""

  def __init__(self, keyname, context):
    self._keyname = keyname
    self._context = context

  def sign(self, instance, doc):
    """return the signed serialization for *instance* in *doc*.

    *doc* is a `lxml` document containing *instance*,
    identified by its id.
    """

    keyname = self._keyname
    if keyname is None: keyname = instance.S_GET_KEYNAME()
    context = self._context
    if context is None: context = default_sign_context
    return context.sign(doc,
                        getattr(instance, instance.S_ID_ATTRIBUTE),
                        keyname
                        )


class SignatureContext(object):
  """encapsulates a signature (sign/verify) context.

  Note: Its instances are sensible. Avoid access from untrusted code!
  """

  def __init__(self, keys_manager=None):
    """initialize a signature context.

    *keys_manager* is currently a mapping from name to a sequence of keys.
    """
    if keys_manager is None: keys_manager = {}
    self.__keys_manager = keys_manager


  def add_key(self, key, name=None):
    """add xmlsec *key* to our context."""
    if name is None: name = key.name
    mngr = self.__keys_manager
    kl = mngr.get(name)
    if kl is None: kl = mngr[name] = []
    kl.append(key)


  def sign(self, doc, id, keyname):
    """sign element identified by *id* in *doc* (`lxml` etree) with the (first) key with *keyname*."""
    from dm.xmlsec.binding import dsig, DSigCtx, findNode, Error
    # will fail with ``IndexError`` when the id does not exist
    node = doc.xpath('//*[@ID="%s"]' % id)[0]
    node = findNode(node, dsig("Signature"))
    assert node is not None, "Missing signature element"
    sign_ctx = DSigCtx(None)
    sign_ctx.signKey = self.__keys_manager[keyname][0]
    try: sign_ctx.sign(node)
    except Error, e:
      raise SignError('signing failed', id, keyname, e)


  def verify(self, doc, id, keyname):
    """verify the node identified by *id* in *doc* using a key associated with *keyname*.

    Raise ``VerifyError``, when the verification fails.

    We only allow a single reference. Its uri must either be empty or
    refer to the element we are verifying.
    In addition, we only allow the standard transforms.
    """
    from dm.xmlsec.binding import findNode, dsig, DSigCtx, \
         TransformInclC14N, TransformExclC14N, TransformSha1, \
         TransformRsaSha1, TransformEnveloped, \
         VerificationError
    node = doc.xpath('//*[@ID="%s"]' % id)
    if len(node) != 1:
      raise VerifyError('id not unique or not found: %s %d' % (id, len(nodes)))
    node = node[0]
    sig = findNode(node, dsig("Signature"))
    # verify the reference.
    refs = sig.xpath('ds:SignedInfo/ds:Reference', namespaces=dsigns)
    if len(refs) != 1:
      raise VerifyError('only a single reference is allowed: %d' % len(refs))
    ref = refs[0]
    uris = ref.xpath('@URI')
    if not uris or uris[0] != '#' + id:
      raise VerifyError('reference uri does not refer to signature parent', id)
    # now verify the signature: try each key in turn
    for key in self.__keys_manager.get(keyname, ()):
      verify_ctx = DSigCtx(None)
      for t in (TransformInclC14N, TransformExclC14N, TransformSha1, TransformRsaSha1):
        verify_ctx.enableSignatureTransform(t)
      for t in (TransformInclC14N, TransformExclC14N, TransformSha1, TransformEnveloped):
        verify_ctx.enableReferenceTransform(t)
      verify_ctx.signKey = key
      try: verify_ctx.verify(sig)
      except VerificationError: pass
      else: return
    raise VerifyError('signature verification failed: %s %s' % (id, keyname))



class _BindingDOMSupportSignatureExtension(object):
  """mixin class to support signatures."""
  _signature_requests = None
  __document = None


  def add_signature_request(self, request, instance):
    """request signature for *instance* in upcoming ``finalize``."""
    sr = self._signature_requests
    if sr is None: sr = self._signature_requests = []
    sr.append((request, instance))


  def finalize(self):
    r = super(_BindingDOMSupportSignatureExtension, self).finalize()
    sr = self._signature_requests
    if sr is None: return r
    # perform all signature requests in reverse order
    from lxml.etree import parse, tostring
    from dm.xmlsec.binding import addIDs
    from StringIO import StringIO
    doc = self.document().toxml()
    doc_tree = parse(StringIO(doc))
    addIDs(doc_tree.getroot(), ['ID'])
    while sr:
      r, i = sr.pop()
      r.sign(i, doc_tree)
    signed_doc = tostring(doc_tree)
    # interestingly: I must use "xml.dom.mindom" for parsing
    #   ``pyxb.utils.domutils.StringToDOM`` does not work
    from xml.dom.minidom import parseString
    self.__document = parseString(signed_doc)
    
  def document(self):
    d = self.__document
    if d is not None: return d
    return super(_BindingDOMSupportSignatureExtension, self).document()


default_sign_context =  SignatureContext()


# Verification support
def verify_signatures(doc, val, keyname=None, context=None):
  """verify all signatures contained in *doc* and update *val*.

  *doc* is a string containing an XML document.
  *val* is its ``pyxb`` binding.

  *keyname*, if given, identifies the key that has verified the enclosing
  context, e.g. the transport.

  *context* is a signature context. Default: ``default_verify_context``

  The function recursively descends *val* and tries to verify
  the signature for each encountered ``Signable`` instance.
  A ``VerifyError`` is raised when the signature verification fails.
  """
  from lxml.etree import parse
  from StringIO import StringIO
  from dm.xmlsec.binding import addIDs

  def verify(node, keyname):
    """verify *node* and (recursively) its decendents."""
    if isinstance(node, Signable):
      # this object is signable
      if getattr(node, node.S_SIGNATURE_ATTRIBUTE) is not None:
        # has its own signature -- verify it
        keyname = node.S_GET_KEYNAME()
        context.verify(dp, getattr(node, node.S_ID_ATTRIBUTE), keyname)
      node.set_signature_verification(keyname)
    # recurse
    ss = getattr(node, '_symbolSet', None)
    if ss is None: return
    # node._symbolSet returns a map eu --> list(child)
    for cl in ss().values():
      for c in cl: verify(c, keyname)

  dp = parse(StringIO(doc))
  if context is None: context = default_verify_context
  addIDs(dp.getroot(), ['ID'])
  verify(val, keyname)


def add_signature_verification():
  """enhance caller globals with signature verification."""
  moddict = get_frame_globals(1)
  cfd = moddict['CreateFromDocument']

  def CreateFromDocument(xml_text, *args, **kw):
    """override to perform signature verification.

    *keyname* and *context* are keyword parameters used for signature
    verification (see ``dm.saml2.signature.verify_signatures``).
    Other arguments are passed on to the original function.
    The keyword parameter *suppress_verification* can suppress the
    verification.
    """
    keyname, context = kw.pop('keyname', None), kw.pop('context', None)
    suppress = kw.pop("suppress_verification", False)
    i = cfd(xml_text, *args, **kw)
    if not suppress: verify_signatures(xml_text, i, keyname, context)
    return i
  CreateFromDocument.signature_enhanced = True

  moddict['CreateFromDocument'] = CreateFromDocument


default_verify_context =  SignatureContext()

dsigns = dict(ds="http://www.w3.org/2000/09/xmldsig#")



# ../../../dm/saml2/pyxb/raw/authncontext.py
# -*- coding: utf-8 -*-
# PyXB bindings for NM:1ec3b375313a629c2fd762e6ebb0af1abc609dd7
# Generated 2012-06-30 20:34:07.140321 by PyXB version 1.1.4
# Namespace urn:oasis:names:tc:SAML:2.0:ac

import pyxb
import pyxb.binding
import pyxb.binding.saxer
import StringIO
import pyxb.utils.utility
import pyxb.utils.domutils
import sys

# Unique identifier for bindings created at the same time
_GenerationUID = pyxb.utils.utility.UniqueIdentifier('urn:uuid:2c59a4e8-c2e2-11e1-9363-0025228242f2')

# Import bindings for namespaces imported into schema


Namespace = pyxb.namespace.NamespaceForURI(u'urn:oasis:names:tc:SAML:2.0:ac', create_if_missing=True)
Namespace.configureCategories(['typeBinding', 'elementBinding'])
ModuleRecord = Namespace.lookupModuleRecordByUID(_GenerationUID, create_if_missing=True)
ModuleRecord._setModule(sys.modules[__name__])

def CreateFromDocument (xml_text, default_namespace=None, location_base=None):
    """Parse the given XML and use the document element to create a
    Python instance.
    
    @kw default_namespace The L{pyxb.Namespace} instance to use as the
    default namespace where there is no default namespace in scope.
    If unspecified or C{None}, the namespace of the module containing
    this function will be used.

    @keyword location_base: An object to be recorded as the base of all
    L{pyxb.utils.utility.Location} instances associated with events and
    objects handled by the parser.  You might pass the URI from which
    the document was obtained.
    """

    if pyxb.XMLStyle_saxer != pyxb._XMLStyle:
        dom = pyxb.utils.domutils.StringToDOM(xml_text)
        return CreateFromDOM(dom.documentElement)
    if default_namespace is None:
        default_namespace = Namespace.fallbackNamespace()
    saxer = pyxb.binding.saxer.make_parser(fallback_namespace=default_namespace, location_base=location_base)
    handler = saxer.getContentHandler()
    saxer.parse(StringIO.StringIO(xml_text))
    instance = handler.rootObject()
    return instance

def CreateFromDOM (node, default_namespace=None):
    """Create a Python instance from the given DOM node.
    The node tag must correspond to an element declaration in this module.

    @deprecated: Forcing use of DOM interface is unnecessary; use L{CreateFromDocument}."""
    if default_namespace is None:
        default_namespace = Namespace.fallbackNamespace()
    return pyxb.binding.basis.element.AnyCreateFromDOM(node, _fallback_namespace=default_namespace)




# Copyright (C) 2010 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Support for the http redirect binding.

The redirect binding allows to transport an SAML2 message
(request or response) via query string parameters in an HTTP ``GET``
request, usually used via an HTTP redirect.

Due to space considerations, the transported message usually
cannot contain a signature. The redirect binding supports
a restricted form of signature for the message's top level.
The implementation here provides the parameters for this signature
support but does not (yet) really implements signing and verification.
Such an implementation requires extension of ``PyXMLsec`` (the
Python binding to the xml security library) or a separate implementation
of the signing and verification algorithms.
"""

from zlib import compress, decompress
from urllib import urlencode
from cgi import parse_qs

from dm.saml2.pyxb.protocol import ResponseType


def encode(url, message, relay_state=None, context=None):
  """encode *message* and *relay_state* as query parameters for *url*; return the extended url.

  *message* is a ``PyXB`` binding object representing either a request or
  response.

  *relay_state* is a string, usually an opaque state handle obtained from
  a relay state manager. In cases, where an unsolicited response is
  sent, *relay_state* can indicate in some way a target.

  *context* is a signature context used for signing (not yet implmented).
  """

  # place of/for '?'
  url = str(url) # might be unicode
  qi = url.find('?')
  if qi < 0: qi = len(url)
  params = dict()
  if relay_state: params['RelayState'] = relay_state
  if message.Signature is not None:
    # Note: we must ensure ``message.Destination`` is set
    message.Destination = url[:qi]
    raise NotImplementedError("signature support not yet implemented")
  # ``compress`` does not really implements the deflate algorithm as of RFC1951
  #   but adds a 2 byte header and a 4 byte trailor (which we remove)
  msg = compress(message.toxml(root_only=True))[2:-4].encode('base64')
  # remove whitespace
  msg = ''.join(msg.split())
  params[isinstance(message, ResponseType) and 'SAMLResponse' or 'SAMLRequest'] \
                             = msg
  qs = urlencode(params)
  if qi == len(url): url += '?' # '?' not yet present
  if qi != len(url)-1 and not url.endswith('&'): url += '&'
  url += urlencode(params)
  return url


def decode(url, context=None):
  """decode *url* and return pair *saml* and *relay_state*.

  *url* is the complete url, including the query string (necessary
  for signature verification).

  *context* is used for signature verification (not yet implemented).
  """
  params = parse_qs(url[url.index('?')+1:]) # maps param to list of values
  if 'Signature' in params:
    raise NotImplementedError('signature support not yet implemented')
  relay_state = params.get('RelayState')
  if relay_state is not None: relay_state = relay_state[0] # unlistify
  msg = (params.get('SAMLRequest') or params.get('SAMLResponse'))[0]
  # Note: RPC 1951 implies an unwrapped deflated data stream
  #  call ``decompress`` with a negative second argument to
  #  specify this case.
  return decompress(msg.decode("base64"), -15), relay_state  




# Copyright (C) 2010 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Binding support."""

SoapBinding = "urn:oasis:names:tc:SAML:2.0:bindings:SOAP"
HttpRedirectBinding = "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"
HttpPostBinding = "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"



from datetime import datetime

from pyxb.bundles.saml20.assertion import *
from dm.saml2.signature import Signable, add_signature_verification
from dm.saml2.uuid import uuid
from dm.saml2.util import enhance, Picklable


class _StandardSaml2Initialization(object):
  """mixin class to provide standard initialization of 'ID', 'Version' and 'IssueInstant."""

  def __init__(self, *args, **kw):
    if 'Version' not in kw: kw['Version'] = '2.0'
    # Note: an id must start with a char
    if 'ID' not in kw: kw['ID'] = '_' + uuid()
    if 'IssueInstant' not in kw: kw['IssueInstant'] = datetime.utcnow()
    super(_StandardSaml2Initialization, self).__init__(*args, **kw)


class _StandardSignable(Signable, _StandardSaml2Initialization):
  """mixin class to provide standard initialization and signability."""

  def S_GET_KEYNAME(self): return self.Issuer.value()


class _AssertionTypeExtension(_StandardSignable, Picklable):
  """Application generated ``Assertion``."""

  def is_valid(self, context=None):
    cs = self.Conditions
    return cs is None or cs.is_valid(context)


enhance('AssertionType')

class ConditionsCheckContext(object):
  """Base class used for condition validity checks.

  Supports the standard condition types.
  """
  #  handling the `Condition` extension point
  #   override as necessary to handle your conditions
  def is_condition_valid(self, condition): return False

  # audience restriction
  # must be overridden by derived classes
  def audience_id(self): return None
  def belongs_to_audience(self, audience):
    return audience == self.audience_id()

  # one time use
  def set_one_time_use(self):
    raise NotImplementedError("we do not support `OneTimeUse`")

  def set_proxy_restriction(self, restriction):
    raise NotImplementedError("we do not support `ProxyRestriction`")


class _ConditionsTypeExtension(object):
  """give it a ``is_valid`` method.

  Concrete implementations likely will need some (potentially) global context.
  """

  def is_valid(self, context=None):
    # according to 2.5.1.1 we should implement a three-value logic
    #  (invalid, valid, indeterminate) -- we fail to do so.
    now = datetime.utcnow()
    cv = self.NotBefore
    if cv is not None and now < cv: return False
    cv = self.NotOnOrAfter
    if cv is not None and now >= cv: return False
    for c in self.Condition:
      if not c.is_valid(context): return False
    for c in self.AudienceRestriction:
      if not c.is_valid(context): return False
    for c in self.OneTimeUse:
      if not c.is_valid(context): return False
    for c in self.ProxyRestriction: pass
    return True

enhance('ConditionsType')


class _ConditionAbstractTypeExtension(object):
  def is_valid(self, context):
    return context is not None and context.is_condition_valid(self)

enhance("ConditionAbstractType")


class _AudienceRestrictionTypeExtension(object):
  def is_valid(self, context):
    if context is None: return False
    for a in self.Audience:
      if context.belongs_to_audience(a): return True
    return False

enhance("AudienceRestrictionType")


class _OneTimeUseTypeExtension(object):
  def is_valid(self, context):
    if context is None: context = ConditionsCheckContext()
    context.set_one_time_use()
    return True

enhance("OneTimeUseType")


class _ProxyRestrictionTypeExtension(object):
  def is_valid(self, context):
    if context is None: context = ConditionsCheckContext()
    context.set_proxy_restriction(self)
    return True

enhance("ProxyRestrictionType")




add_signature_verification()



from os.path import abspath, dirname, join
try:
  # try to use setuptools
  from setuptools import setup
  setupArgs = dict(
      include_package_data=True,
      install_requires=[
        'setuptools', # to make `buildout` happy
        'dm.xmlsec.binding',
        'pyxb >= 1.1.4',
      ] ,
      namespace_packages=['dm',
                          ],
      zip_safe=False,
      entry_points = dict(
        ),
      )
except ImportError:
  # use distutils
  from distutils import setup
  setupArgs = dict(
    )

cd = abspath(dirname(__file__))
pd = join(cd, 'dm', 'saml2')

def pread(filename, base=pd): return open(join(base, filename)).read().rstrip()


setup(name='dm.saml2',
      version=pread('VERSION.txt').split('\n')[0],
      description="SAML2 support based on PyXB",
      long_description=pread('README.txt'),
      classifiers=[
        # 'Development Status :: 3 - Alpha',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.4',
        'Programming Language :: Python :: 2.6',
        'Topic :: Utilities',
        ],
      author='Dieter Maurer',
      author_email='dieter@handshake.de',
      url='http://pypi.python.org/pypi/dm.xmlsec.pyxb',
      packages=['dm', 'dm.saml2', 'dm.saml2.pyxb'],
      license='BSD',
      keywords='saml2 pyxb',
      **setupArgs
      )



# Copyright (C) 2010-2012 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Elementary metadata support.

We have only limited support for signing: our signature mini-framework
requires that the keyname must be derived from the signed element. Not
all defined elements support this derivation. We support signing
based on ``entityID`` for elements that have this.

Due to this restriction, it is difficult to exchange signature protected
new signature key information. The new information would need to be signed
with the old key, which may not be easy for the signer. Probably,
we must extend our signing framework and look for certificates
provided in ``KeyInfo`` elements.
"""


from dm.saml2.uuid import uuid
from dm.saml2.util import enhance, Picklable
from dm.saml2.signature import Signable, add_signature_verification

from pyxb.bundles.saml20.metadata import *


class _StandardSaml2Initialization(object):
  """mixin class to provide standard initialization of 'ID'."""

  def __init__(self, *args, **kw):
    if 'ID' not in kw: kw['ID'] = '_' + uuid()
    super(_StandardSaml2Initialization, self).__init__(*args, **kw)


class _StandardSignable(Signable, _StandardSaml2Initialization):
  """mixin class to provide standard initialization and signability."""

  def S_GET_KEYNAME(self): return self.entityID


class _EntityDescriptorTypeExtension(_StandardSignable, Picklable): pass

enhance('EntityDescriptorType')


class _RoleDescriptorTypeExtension(object):
  """mixin class to provide standard initialization of ``protocolSupportEnumeration``."""
  def __init__(self, *args, **kw):
    pse = kw.get('protocolSupportEnumeration')
    proto = 'urn:oasis:names:tc:SAML:2.0:protocol'
    if pse is None: pse = kw['protocolSupportEnumeration'] = proto
    if proto not in pse: kw['protocolSupportEnumeration'] += ' ' + proto
    super(_RoleDescriptorTypeExtension, self).__init__(*args, **kw)
  
enhance('IDPSSODescriptorType', _RoleDescriptorTypeExtension)
enhance('SPSSODescriptorType', _RoleDescriptorTypeExtension)
enhance('AuthnAuthorityDescriptorType', _RoleDescriptorTypeExtension)
enhance('AttributeAuthorityDescriptorType', _RoleDescriptorTypeExtension)
enhance('PDPDescriptorType', _RoleDescriptorTypeExtension)


add_signature_verification()



# Copyright (C) 2010-2012 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Metadata support.

For most SAML2 operations metadata describing important aspects
of a set of SAML2 authorities is vital.

This module defines base classes to manage and access SAML2 related
metadata.

At its center is the ``MetadataRepository``. It behaves like
a container for entity references, called ``Entity``.
Each entity reference corresponds to an SAML2 authority and
knows how to obtain metadata relevant for this authority.
Thus, the entity reference encapsulates the necessary configuration
to access its authorities metadata.

A ``MetadataRepository`` can be asked for current metadata for
its known authorities; access can be either via entity id or its SHA-1
hash (used by SAML2 for some purposes).

Metadata for a specific entity is described by the class
``EntityMetadata``. This class can handle multiple metadata sets
for a single entity valid for different time ranges. This
complexity was mainly introduced to facilitate "key rollover".
The class knows how to fetch new metadata (via use of ``Entity``)
and can continue to use outdated metadata when this fails.

For efficiency reasons, ``MetadataRepository`` uses an externally
controllable store, used by ``EntityMetadata`` for its
storage needs. It must be able to store arbitrary mappings
from ``unicode`` to picklable objects. If this store is persistent,
metadata might be retained across process lifetimes (depends
on other requirements as well).
"""
from logging import getLogger
from datetime import datetime, timedelta
from urllib2 import urlopen


logger = getLogger(__name__)


class MetadataFetchError(Exception):
  """Exception to report the failure fetching metadata.

  Its constructor expects two arguments, the entity id
  and an exception.
  """
  def __repr__(self):
    id, exc = self.args
    return 'MetadataFetchError for %s: %s' % (id, exc)

  __str__ = __repr__


class EntityBase(object):
  """abstract entity reference base class.

  Encapsulates the relevant configuration data to identify
  an SAML2 authority and access its current metadata description.
  """

  id = None # entity id, definied in ``__init__``

  # "dm.saml2" assumes specific keys (rather than certificate chains founded
  #   in a set of trusted certificates with subsequent subject verification)
  #   for its verifications.
  #   Usually, such keys would be made known via metadata exchange.
  #   This obviously poses a chicken egg problem when signatures
  #   must be verified for a metadata exchange itself.
  #   The following attribute allows to suppress the signature check
  suppress_verification = True

  def __init__(self, id, *args, **kw):
    # id's are hashed; thus, they should be bytes (not unicode)
    super(EntityBase, self).__init__(*args, **kw)
    self.id = str(id)

  def get_metadata(self):
    """fetch metadata for this entity (as Python object)."""

    from dm.saml2.pyxb.metadata import CreateFromDocument
    # Note: the unrestricted "try: ... except: ..." is nasty
    #  but there are too many potential exceptions that
    #  an enumeration would be difficult
    try:
      return CreateFromDocument(
        self.get_metadata_document(),
        suppress_verification=self.suppress_verification,
        )
    except Exception, e:
      logger.exception('failed to get metadata for %s', self.id)
      raise MetadataFetchError(self.id, e)

  def get_metadata_document(self):
    """fetch metadata for this entity (as XML document)."""
    raise NotImplementedError('``get_metadata_document`` must be implemended by derived classes')


class EntityByUrl(EntityBase):
  """Entity reference obtaining metadata from a well known internet url.

  This is a simple class. We might want significantly more transport control,
  e.g. proxy, authentication, ...
  """

  def __init__(self, id, url, *args, **kw):
    super(EntityByUrl, self).__init__(id, *args, **kw)
    self.url = url

  def get_metadata_document(self):
    return urlopen(self.url).read()



class EntityMetadata(object):
  """Entity proxy providing access to the entities metadata.

  The class manages a sequence of metadata sets -- essentially
  to facilitate key rollover.

  The class automatically tries to fetch new metadata when
  existing metadata lost its validity. However, it is ready
  to continue to use outdated metadata when fetching a new one fails.

  Under normal conditions, it maintains a single metadata set.
  However, ``fetch_metadata`` can be used, to fetch a current metadata set,
  even when there are valid ones. This may facilitate key rollover.

  The class uses an externally provided *store* (a mapping) to
  store its metadata.
  """
  # to be overridden by derived classes (to facilitate persistency)
  SEQUENCE_CLASS = list


  def __init__(self, entity, store, default_validity):
    self.entity, self.store, self.default_validity = entity, store, default_validity
    self.id = entity.id


  def _get_metadata_sets(self, exc=False):
    """the currently known metadata sets;

    *exc* requests an exception when no set is available.
    """
    store = self.store; eid = self.id
    modified = False
    sets = store.get(eid)
    if sets is None:
      sets = self.SEQUENCE_CLASS(); modified = True
    # find sets no longer valid
    #  Note: sets is a sequence, ordered by "most recent" first
    now = datetime.now()
    invalid = [] # list of indexes for no longer valid entries
    for i, s in enumerate(sets):
      if now > s.validUntil: invalid.append(i)
    if len(invalid) == len(sets) or not len(sets):
      # no valid sets -- try to fetch a new one
      s = self._fetch(exc)
      if s is not None: sets[:] = (s,)
      else:
        # try to keep the most recent one
        sets = sets[:1]
      modified = True
    elif len(invalid):
      # invalidate no longer valid entries
      invalid.reverse()
      for i in invalid: del sets[i]
      modified = True

    # support for ZODB persistency
    if modified: store[eid] = sets
    return modified, sets

  def get_metadata_sets(self, exc=False):
    return self._get_metadata_sets(exc)[1]


  def get_recent_metadata(self):
    # can be optimized
    return self.get_metadata_sets(True)[0]


  def fetch_metadata(self, clear_old=False):
    """fetch a new metadata set.

    Without explicit calls of this method, we have at most 1 set
    as we fetch a new set only when all others became invalid.

    The method would be called when an entity reports changes
    to its metadata.

    Without *clear_old*, certificates defined by older sets
    will still be used (to facilitate key rollover).
    """
    sets = self.get_metadata_sets()
    sets.insert(0, self._fetch())
    if clear_old: del sets[1:]
    self.store[self.id] = sets


  def clear_metadata(self):
    """delete the stored metadata."""
    self.store[self.id] = self.SEQUENCE_CLASS()


  def get_role_descriptor(self, role,
                          protocol='urn:oasis:names:tc:SAML:2.0:protocol'
                          ):
    """the (first) role descriptor supporting *role* and *protocol*, or ``None``.

    According to the spec, their should be at most one such descriptor.

    Roles are represented by short strings: ``idpsso`` (identity provider),
    ``spsso`` (service provider), ``authn`` (authentication authority),
    ``attribute`` (attribute authority) and ``pdp`` (policy decision
    point).
    """
    return select_role_descriptor(self.get_recent_metadata(),
                                  role, protocol
                                  )

  def get_role_certificates(self, role,
                            use=None,
                            protocol='urn:oasis:names:tc:SAML:2.0:protocol'
                            ):
    """select all certificates used by *role*.

    To facilitate key rollover, older metadata sets are used as well
    to determine keys.

    See ``get_role_descriptor`` for the coding of *roles*.
    """
    certs = []
    for s in self.get_metadata_sets():
      select_role_certificates(select_role_descriptor(s, role, protocol),
                               use, certs
                               )
    return certs

  def _fetch(self, exc=True):
    """fetch current metadata.

    If *exc*, raise exceptions when the metadata cannot be fetched;
    otherwise, return ``None``.
    """
    eid = self.id
    try: s = self.entity.get_metadata()
    except MetadataFetchError:
      if exc: raise
      return
    else:
      # fetch successfull
      if eid != s.entityID:
        # looks like metadata describing someone else
        if not exc: return
        raise MetadataFetchError(
          eid,
          ValueError('entityID mismatch: %s != %s' % (eid, s.entityID))
          )
      # we do not check that metadata is authentic

      # the standard requires that either ``validUntil`` or ``cacheDuration``
      #   is defined -- but not all SAML2 entities follow the standard
      #   normalize to always provide ``validUntil``
      if s.validUntil is None:
        cd = s.cacheDuration or self.default_validity
        s.validUntil = datetime.now() + cd
    return s


role2element = dict(
  idpsso='IDPSSODescriptor',
  spsso='SPSSODescriptor',
  authn='AuthnAuthorityDescriptor',
  ap='AttributeAuthorityDescriptor',
  pdp='PDPDescriptor',
  )

def select_role_descriptor(md, role, protocol):
  """select the first descriptor from metadata *md* supporting *role* and *protocol*.

  See ``EntityMetadata.get_role_descriptor`` for further details.
  """
  for r in getattr(md, role2element[role]):
    if protocol in r.protocolSupportEnumeration: return r


def select_role_certificates(rd, use, certs):
  """extend *certs* (a list) by certificates used by role descriptor *rd*.

  Note: *rd* may be ``None``.
  """
  if rd is None: return
  for kd in rd.KeyDescriptor:
    if use is None or kd.use is None or use == kd.use:
      ki = kd.KeyInfo
      for x509 in getattr(ki, 'X509Data', ()):
        cert = getattr(x509, 'X509Certificate', None)
        if cert is not None:
          # `cert` contains a certificate chain. We use only the first one
          certs.append(cert[0])


class MetadataRepository(object):
  """``Entity`` container managing a set of entity references.

  Provides access to ``EntityMetadata``, either by id or by its SHA-1 hash.
  """

  # can be overridden by derived classes -- e.g. to provide persistency
  INTERNAL_STORAGE_CLASS = dict
  METADATA_STORAGE_CLASS = dict
  METADATA_CLASS = EntityMetadata

  default_validity = timedelta(10, 0, 0) # 10 days

  def __init__(self, default_validity=None):
    self._entities = self.INTERNAL_STORAGE_CLASS()
    self._metadata = self.INTERNAL_STORAGE_CLASS()
    self._metadata_storage = self.METADATA_STORAGE_CLASS()
    if default_validity is not None: self.default_validity = default_validity
  

  def add_entity(self, entity):
    """add new *entity*."""
    ents = self._entities; eid = entity.id
    if eid in ents: raise ValueError('entity already managed: %s' % id)
    self._entities[eid] = entity
    self._entities = self._entities # to facilitate ZODB persistency

  def list_entities(self):
    # Note: when entities are modified, it might become necessary
    #  to invalidate the associated metadata
    return self._entities.values() # should we sort them?

  def get_entity(self, id, default=None):
    """the entity *id* or *default*."""
    return self._entities.get(id, default)

  def del_entity(self, eid):
    """delete entity identified by *eid* (if present)."""
    if eid in self._entities: del self._entities[eid]
    if eid in self._metadata: del self._metadata[eid]

  def metadata_by_id(self, eid):
    md = self._metadata.get(eid)
    if md is None:
      md = self._metadata[eid] = self.METADATA_CLASS(
        self._entities[eid], self._metadata_storage, self.default_validity
        )
      self._metadata = self._metadata # to facilitate ZODB persistency
    return md


  def metadata_by_hash(self, sha_hash):
    h2id = self._get_hash2id()
    if sha_hash not in h2id: h2id = self._get_hash2id(True)
    return metadata_by_id(h2id[sha_hash])


  _hash2id = None

  def _get_hash2id(self, rebuild=False):
    h2id = self._hash2id
    if h2id is None or rebuild:
      from sha import sha
      h2id = self._hash2id = dict((sha(eid), eid) for eid in self._entities)
    return h2id



# Copyright (C) 2010 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Support for the http post binding.

The post binding allows to transport an SAML2 message
(request or response) via parameters in an HTTP ``POST``
request, usually used via an HTML form submission, either explicitely
or automatically via Javascript.
"""
from cgi import escape

from dm.saml2.pyxb.protocol import ResponseType


def encode(message, relay_state=None):
  """encode *message* and *relay_state*, returning a dict of parameters.

  *message* is a ``PyXB`` binding for an SAML request or response.
  """
  params = dict()
  if relay_state is not None: params['RelayState'] = relay_state
  params[isinstance(message, ResponseType) and 'SAMLResponse' or 'SAMLRequest'] = \
                             message.toxml().encode('base64')
  return params


def as_controls(params):
  """encode *params* (a dict) a hidden form controls (an XHTML fragment)."""
  cl = []
  for pn, pv in params.iteritems():
    cl.append('<input name="%s" type="hidden" value="%s" />'
              % (escape(pn, True), escape(pv, True))
              )
  return ''.join(cl)


def as_form(params, action, submit=None, name=None):
  """encode *params* (a dict) as a simple form (an XHTML fragment).

  *action* (an URL) specifies the form action. 

  *submit* specifies the submit button label. It usually will be a
  unicode value. In this case, the result will be unicode as well.

  *name* can be used to give the form a name.
  """
  name = name and 'name="%s"' % escape(name, True) or ''
  sv = submit and 'value="%s"' % escape(submit, True) or ''
  return (
    '<form %s action="%s" method="POST">%s<input type="submit" %s /></form>'
    % (name, action, as_controls(params), sv)
    )


def decode(params):
  """decode (dict) *params* into a pair *saml*, *relay_state*.

  ``decode`` is (almost) the inverse of ``encode`` (it returns
  an saml document rather than a `PyXB` object.
  """
  return ((params.get('SAMLRequest') or params.get('SAMLResponse'))
          .decode('base64'),
          params.get('RelayState')
          )






# Copyright (C) 2010-2012 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Utilities"""
from sys import _getframe, modules

def get_frame_globals(level):
  """the globals at frame level *level* above us."""
  fr = _getframe(level + 1)
  return fr.f_globals


def enhance(name, extension=None):
  r"""look up *name* in the callers globals and enhance it by *extension*.

  If *extension* is not specified, it is looked up in the callers
  globals under ``_``\ *name*\ ``Extension``.

  The enhanced object is supposed to behave similar to a class.
  Enhancement happens by defining a new class with *extension*
  as first ``__bases__`` element.
  """
  mdict = get_frame_globals(1)
  if extension is None: extension = mdict['_%sExtension' % name]
  base = mdict[name]
  ec = mdict[name] = type(base)(name, (extension, base) + base.__bases__, {})
  ec.__module__ = mdict["__name__"] # ensure picklability
  base._SetSupersedingClass(ec)


class Picklable(object):
  """Mixin class to implement pickling via XML serialization.

  Note: PyXB objects should be picklable by itself, but apparently
  some literals (at least `dateTime`) have problems.

  Note: Our pickling is not completely faithfull. We loose signature
  verification information (but this could be fixed, if necessary).
  """
  def __getstate__(self): return self.toxml(root_only=True)
  def __setstate__(self, state):
    mod = modules[self.__class__.__module__]
    cfd = mod.CreateFromDocument
    if getattr(cfd, "signature_enhanced", False):
      obj = cfd(state, suppress_verification=True)
    else: obj = cfd(state)
    self.__dict__.update(obj.__dict__) # quite brutal


# Auxiliary functions to better access PyXB objects.
def children(obj, none_ok=False):
  """the children of *obj* as sequence of `ElementUse`, value pairs.

  If *none_ok*, all children are reported; otherwise, only children
  with a value different from None.
  """
  return [(eu, eu.value(obj))
          for eu in obj._ElementMap.values()
          if none_ok or eu.value(obj) is not None
          ]

def child_items(obj, none_ok=False):
  """the children of *obj* as sequence of attr, value pairs."""
  return [(eu.id(), value) for (eu, value) in children(obj, none_ok)]

def child_values(obj, none_ok=False):
  """the children of *obj* as sequence of values."""
  return [ci[1] for ci in children(obj, none_ok)]

def child_ids(obj, none_ok=False):
  """the children of *obj* as sequence of ids."""
  return [ci[0].id() for ci in children(obj, none_ok)]



##############################################################################
## Authentication class support

SAML_CLASS_PREFIX = "urn:oasis:names:tc:SAML:2.0:ac:classes:"

def normalize_class(c):
  """normalize authentication context class reference *c*."""
  if ":" not in c: c = SAML_CLASS_PREFIX + c
  return c

# the authentication classes supported so far
SAML_CLASSES = (
  normalize_class("Password"),
  normalize_class("PasswordProtectedTransport"),
  )

SAML_CLASS_STRENGTH = {
  normalize_class("Password") : 100,
  normalize_class("PasswordProtectedTransport") : 200,
  }

def compare_classes(c1, c2):
  """compare the strength of classes *c1* and *c2*.

  return None if this is impossible; otherwise, -1, 0 or 1.
  """
  if c1 == c2: return 0
  if c1 not in SAML_CLASS_STRENGTH or c2 not in SAML_CLASS_STRENGTH: return
  return cmp(SAML_CLASS_STRENGTH[c1], SAML_CLASS_STRENGTH[c2])



##############################################################################
## Name id support
##
##  In an ealier version, some nameid formats used a wrong prefix.
##  Thanks to Dylan Jay for pointing out the problem and
##  helping to fix it.
NAMEID_FORMAT_PREFIX = "urn:oasis:names:tc:SAML:2.0:nameid-format:"
NAMEID_FORMAT_PREFIX1 = "urn:oasis:names:tc:SAML:1.1:nameid-format:"

SAML11_NAMEID_FORMATS = frozenset((
  "unspecified", "emailAddress", "X509SubjectName", 
  "WindowsDomainQualifiedName",
  ))


def normalize_nameid_format(nidf):
  """normalize nameid format *nidf*."""
  if ":" not in nidf:
    nidf = (nidf in SAML11_NAMEID_FORMATS and NAMEID_FORMAT_PREFIX1 or NAMEID_FORMAT_PREFIX) + nidf
  return nidf

NAMEID_FORMATS = (
  normalize_nameid_format("emailAddress"),
  normalize_nameid_format("X509SubjectName"),
  normalize_nameid_format("WindowsDomainQualifiedName"),
  normalize_nameid_format("kerberos"),
  normalize_nameid_format("entity"),
  normalize_nameid_format("persistent"),
  normalize_nameid_format("transient"),
  # moved to the end of the list, to give more specific formats better exposure
  normalize_nameid_format("unspecified"),
  # not a first class nameid format -- usuably only as part of `NameIDPolicy`
  #normalize_nameid_format("encrypted"),
  )


##############################################################################
## attribute support

ATTRNAME_FORMAT_PREFIX = "urn:oasis:names:tc:SAML:2.0:attrname-format:"

def normalize_attrname_format(anf):
  """normalize attribute name format *anf*."""
  if ":" not in anf: anf = ATTRNAME_FORMAT_PREFIX + anf
  return anf

ATTRNAME_FORMATS = (
  normalize_attrname_format("unspecified"),
  normalize_attrname_format("uri"),
  normalize_attrname_format("basic"),
  )


##############################################################################
## XML-schema basic type conversion

# impossible due to local/global import confusion
#import pyxb.binding.datatypes as xs
xs = __import__("pyxb.binding.datatypes", {}, {}, ["__name__"])

XSCHEMA_BASE_TYPES = (
  "anyURI", "base64Binary", "boolean", "date", "dateTime", "double",
  "duration", "float", "int", "string",
  )

def xs_convert_to_xml(ty, value, element=None):
  """convert Python *value* to XML using base type *ty*.

  If *element* put into such an element -- otherwise return the
  PyXB basic binding of *value* corresponding to *ty*.

  If *value* is a list or tuple, convert its elements and return
  a list.
  """
  if value is None: return element is not None and element() or None
  if isinstance(value, (list, tuple)):
    # Note: we assume to have only one sequence level
    return [xs_convert_to_xml(ty, v, element) for v in value]
  return getattr(xs, ty)(value, _element=element)

def xs_convert_from_xml(xml, is_seq=False):
  """convert *xml* to Python.

  *xml* is the PyXB binding of an element with simple content.

  As XML-Schema often uses lists to represent optional values,
  *is_seq* controls whether the result should be a list or
  whether the convertion should try to convert to a single value
  (`None`, if the list is empty; the first component if the list has
  a single value).
  """
  if xml is None: return
  if is_seq:
    # *xml* must be a sequence, too; crash, if this is not the case
    return [xs_convert_from_xml(e) for e in xml]
  if isinstance(xml, (list, tuple)):
    if not xml: return
    if len(xml) == 1: return xs_convert_from_xml(xml[0])
    raise TypeError("cannot convert list to single value")
  # most PyXB types are perfect Python values -- only a few types require conversion
  if isinstance(xml, xs.dateTime): return pyxb_to_datetime(xml)
  # some types may still be missing
  return xml



##############################################################################
## Wrapper

class Printer(object):
  """Wrapper to print a PyXB object."""

  def __init__(self, obj): self._obj = obj

  def __str__(self): return self._obj.toxml(root_only=True)
  __repr__ = __str__



##############################################################################
## Workaround for PyXB weaknesses

def pyxb_to_datetime(pdt):
  """PyXB datetime values are not picklable -- convert to "normal" `datetime`."""
  if pdt is None: return
  from datetime import datetime
  return datetime(pdt.year, pdt.month, pdt.day, pdt.hour, pdt.minute, pdt.second, pdt.microsecond)
